import time
import hashlib
import hmac
from json import dumps

api_key = 'usergeneratedoNtGtKG0YW6sLSj1Q12Q7hOCVhTAgvd0xMexdLK4qZZhnQbSO1'


def update_hmac(hmac_obj: hmac.HMAC, update_items: list) -> hmac.HMAC:
    """Add update_items to hmac_obj and return hmac_obj."""
    print(f"HMAC Hexdigest (base):        {hmac_obj.hexdigest()}")
    for item in update_items:
        hmac_obj.update(item.encode())
        print
    return hmac_obj


def sign(method: str, endpoint: str, payload: dict = None, timestamp: str = None,):
    """HMAC signing function for Mhanndalorian Bot APIs."""
    if timestamp:
        req_time = timestamp
    else:
        req_time = str(int(time.time() * 1000))
    print(f"{req_time=}")
    hmac_obj = hmac.new(key=api_key.encode(), digestmod=hashlib.sha256)
    print(f"HMAC Hexdigest (base):        {hmac_obj.hexdigest()}")
    hmac_obj.update(req_time.encode())
    print(f"HMAC Hexdigest (timestamp):   {hmac_obj.hexdigest()}")
    hmac_obj.update(method.upper().encode())
    print(f"HMAC Hexdigest (HTTP method): {hmac_obj.hexdigest()}")
    hmac_obj.update(endpoint.lower().encode())
    print(f"HMAC Hexdigest (endpoint):    {hmac_obj.hexdigest()}")
    payload_str = dumps(payload or {"payload": {"allyCode": "314927874"}}, separators=(',', ':'))
    print(f"{payload_str=}")
    payload_hash = hashlib.md5(payload_str.encode()).hexdigest()
    print(f"{payload_hash=}")
    hmac_obj.update(payload_hash.encode())
    return {"x-timestamp": req_time, "Authorization": hmac_obj.hexdigest()}


def hmac_sign(api_key: str, method: str, endpoint: str, payload: dict) -> dict[str, str]:
    """HMAC signing function for Mhanndalorian Bot APIs.
    
        Args
            api_key: The API key assigned by Mhanndalorian Bot
            method: The HTTP method of the request
            endpoint: The API endpoint URI
            payload: Dictionary containing the payload of the HTTP request

        Returns
            Dictionary containing the 'x-timestamp' and 'Authorization' headers for the signed request
    """
    # Get the current time in milliseconds
    req_time = str(int(time.time() * 1000))

    # Create a base HMAC object using SHA256 algorithm
    hmac_obj = hmac.new(key=api_key.encode(), digestmod=hashlib.sha256)

    # Add the request timestamp to the HMAC object
    hmac_obj.update(req_time.encode())

    # Add the HTTP method (in upper case) to the HMAC object
    hmac_obj.update(method.upper().encode())

    # Add the API endpoint URI to the HMAC object
    hmac_obj.update(endpoint.lower().encode())

    # Create a serialized string from the payload JSON/dictionary object
    payload_str = dumps(payload, separators=(',', ':'))

    # Generate MD5 hash of the payload string
    payload_hash = hashlib.md5(payload_str.encode()).hexdigest()

    # Add the payload MD5 hash to the HMAC object
    hmac_obj.update(payload_hash.encode())

    # Return the HTTP headers that must be included with the HMAC signed request
    return {"x-timestamp": req_time, "Authorization": hmac_obj.hexdigest()}



import pytest
from mhanndalorian_bot.base import MBot


def test_cleanse_allycode_valid():
    """Test valid allycode cleansing."""
    result = MBot.cleanse_allycode("123-456-789")
    assert result == "123456789"


def test_cleanse_allycode_invalid_length():
    """Test invalid allycode with incorrect length."""
    with pytest.raises(ValueError, match="Value must be exactly 9 numerical characters."):
        MBot.cleanse_allycode("12345678")


def test_cleanse_allycode_with_non_digit_characters():
    """Test invalid allycode with non-digit characters."""
    with pytest.raises(ValueError, match="Value must be exactly 9 numerical characters."):
        MBot.cleanse_allycode("12345678a")


def test_cleanse_discord_id_valid():
    """Test valid Discord ID cleansing."""
    result = MBot.cleanse_discord_id("123456789012345678")
    assert result == "123456789012345678"


def test_cleanse_discord_id_invalid_length():
    """Test invalid Discord ID with incorrect length."""
    with pytest.raises(ValueError, match="Value must be exactly 18 numerical characters."):
        MBot.cleanse_discord_id("12345678")


def test_cleanse_discord_id_with_non_digit_characters():
    """Test invalid Discord ID with non-digit characters."""
    with pytest.raises(ValueError, match="Value must be exactly 18 numerical characters."):
        MBot.cleanse_discord_id("12345678901234567a")


def test_set_api_key_valid():
    """Test setting a valid API key."""
    bot = MBot(api_key="12345678abcdefgh", allycode="123456789")
    bot.set_api_key("newapikey")
    assert bot.api_key == "newapikey"
    assert bot.headers["api-key"] == "newapikey"


def test_set_api_key_invalid_type():
    """Test setting an invalid API key with a non-string type."""
    bot = MBot(api_key="12345678abcdefgh", allycode="123456789")
    with pytest.raises(ValueError, match="api_key must be a string"):
        bot.set_api_key(12345678)


def test_set_allycode_valid():
    """Test setting a valid allycode."""
    bot = MBot(api_key="12345678abcdefgh", allycode="123456789")
    bot.set_allycode("987-654-321")
    assert bot.allycode == "987654321"
    assert bot.payload["payload"]["allyCode"] == "987654321"


def test_set_allycode_invalid():
    """Test setting an invalid allycode."""
    bot = MBot(api_key="12345678abcdefgh", allycode="123456789")
    with pytest.raises(ValueError, match="Value must be exactly 9 numerical characters."):
        bot.set_allycode("98765")


def test_set_api_host_valid():
    """Test setting a valid API host."""
    MBot.set_api_host("https://testhost.com")
    assert MBot.api_host == "https://testhost.com"


def test_set_api_host_invalid_type():
    """Test setting an invalid API host with a non-string type."""
    with pytest.raises(ValueError, match="api_host must be a string"):
        MBot.set_api_host(12345)



# coding=utf-8
"""
Class definition for SWGOH MHanndalorian Bot API module
"""

from __future__ import absolute_import, annotations

from typing import Any, Dict, Optional

from .base import EndPoint, MBot


class API(MBot):
    """
    Container class for MBot module to facilitate interacting with Mhanndalorian Bot authenticated
    endpoints for SWGOH. See https://mhanndalorianbot.work/api.html for more information.
    """

    def fetch_data(self, endpoint: str | EndPoint,
                   *, method: Optional[str] = None, hmac: Optional[bool] = None) -> Dict[Any, Any]:
        """Return data from the provided API endpoint using standard synchronous HTTP requests

            Args
                endpoint: API endpoint as a string or EndPoint enum

            Keyword Args
                method: HTTP method as a string, defaults to POST
                hmac: Boolean flag indicating whether the endpoints requires HMAC signature authentication

            Returns
                Dictionary from JSON response, if found.
        """
        if isinstance(endpoint, EndPoint):
            endpoint = f"/api/{endpoint.value}"

        method = method.upper() if method else "POST"

        if isinstance(hmac, bool):
            signed = hmac
        else:
            signed = self.hmac

        self.logger.debug(f"Endpoint: {endpoint}, Method: {method},  HMAC: {signed}")

        if signed:
            self.logger.debug(f"Calling HMAC signing method ...")
            self.sign(method=method, endpoint=endpoint, payload=self.payload)

        result = self.client.post(endpoint, json=self.payload)

        self.logger.debug(f"HTTP request headers: {result.request.headers}")
        self.logger.debug(f"API instance headers attribute: {self.headers}")

        if result.status_code == 200:
            return result.json()
        else:
            return {"msg": "Unexpected result", "reason": result.content.decode()}

    async def fetch_data_async(self, endpoint: str | EndPoint,
                               *, method: Optional[str] = None, hmac: Optional[bool] = None) -> Dict[Any, Any]:
        """Return data from the provided API endpoint using asynchronous HTTP requests

            Args
                endpoint: API endpoint as a string or EndPoint enum

            Keyword Args
                method: HTTP method as a string, defaults to POST
                hmac: Boolean flag indicating whether the endpoints requires HMAC signature authentication

            Returns
                httpx.Response object
        """
        if isinstance(endpoint, EndPoint):
            endpoint = f"/api/{endpoint.value}"

        method = method.upper() if method else "POST"

        if isinstance(hmac, bool):
            signed = hmac
        else:
            signed = self.hmac

        if signed:
            self.sign(method=method, endpoint=endpoint, payload=self.payload)

        result = await self.aclient.post(endpoint, json=self.payload)

        if result.status_code == 200:
            return result.json()
        else:
            return {"msg": "Unexpected result", "reason": result.content.decode()}



"""
Base class definition for core object for storing and sharing information for use by all modules
"""

from __future__ import absolute_import, annotations

import hashlib
import hmac
import logging
import time
from json import dumps
from typing import Any, Optional

import httpx
from sentinels import Sentinel

from .attrs import APIKey, AllyCode, Debug, EndPoint, HMAC, Headers, Payload

# Define sentinels used in parameter checking
NotSet = Sentinel('NotSet')


class MBot:
    """Base class for MBot modules

    Args
        api_key: MHanndalorian Bot API key as a string
        allycode: Player allycode as a string

    Keyword Args
        api_host: Optional host URL for MHanndalorian Bot API, defaults to https://mhanndalorianbot.work/
        hmac: Boolean flag indicating whether the endpoints should use HMAC signature authentication, Default: True
    """

    api_host: str = "https://mhanndalorianbot.work"
    logger: logging.Logger = logging.getLogger("mbot")

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)
    logger.addHandler(console_handler)

    debug = Debug(False)
    hmac = HMAC(True)
    api_key: str = APIKey()
    allycode: str = AllyCode()

    client: httpx.Client = httpx.Client(base_url=f"{api_host}", timeout=15, verify=False)
    aclient: httpx.AsyncClient = httpx.AsyncClient(base_url=f"{api_host}", timeout=15, verify=False)

    def __init__(self, api_key: str, allycode: str, *, api_host: Optional[str] = None, hmac: Optional[bool] = True):

        self.api_key = api_key
        self.allycode = allycode
        self.headers = Headers({"Content-Type": "application/json", "api-key": self.api_key})
        self.payload = Payload({"payload": {"allyCode": self.allycode}})

        # self.set_api_key(api_key)
        # self.set_allycode(allycode)

        if isinstance(api_host, str):
            self.set_api_host(api_host)

        if isinstance(hmac, bool):
            self.hmac = hmac

    @staticmethod
    def cleanse_allycode(allycode: str) -> str:
        """Remove any dashes from provided string and verify the result contains exactly 9 digits"""
        if not isinstance(allycode, str):
            raise ValueError(f"{allycode} must be a string, not type:{type(allycode)}")

        allycode = allycode.replace('-', '') if '-' in allycode else allycode

        if not allycode.isdigit() or len(allycode) != 9:
            raise ValueError(f"Invalid allyCode ({allycode}): Value must be exactly 9 numerical characters.")

        return allycode

    @staticmethod
    def cleanse_discord_id(discord_id: str) -> str:
        """Validate that discord ID is an 18 character string of only numerical digits"""
        if not isinstance(discord_id, str):
            raise ValueError(f"{discord_id} must be a string, not type: {type(discord_id)}")

        if not discord_id.isdigit() or len(discord_id) != 18:
            raise ValueError(f"Invalid Discord ID ({discord_id}): Value must be exactly 18 numerical characters.")

        return discord_id

    def get_api_key(self):
        """Return masked API key for logging purposes."""
        return f"{'*' * 6 + self.api_key[-8:]}"

    def set_api_key(self, api_key: str):
        """Set the api_key value for the container class and update relevant attributes (including headers)"""

        if not isinstance(api_key, str):
            raise ValueError("api_key must be a string")

        setattr(self, "api_key", api_key)

        self.headers["api-key"] = self.api_key
        self.payload["payload"]["allyCode"] = self.allycode
        self.client.headers = self.headers
        self.aclient.headers = self.headers

    def set_allycode(self, allycode: str):
        """Set the allycode value for the container class and update relevant attributes"""

        allycode = self.cleanse_allycode(allycode)

        setattr(self, "allycode", allycode)

        self.payload["payload"]["allyCode"] = allycode

    @classmethod
    def set_api_host(cls, api_host: str):
        """Set the api_host value for the container class and update relevant attributes"""

        if not isinstance(api_host, str):
            raise ValueError("api_host must be a string")

        setattr(cls, "api_host", api_host)

        cls.client.base_url = f"{api_host}/api/"
        cls.aclient.base_url = f"{api_host}/api/"

    @classmethod
    def set_client(cls, **kwargs: Any):
        """Set the client values for the container class and update relevant attributes"""
        for key, value in kwargs.items():
            setattr(cls.client, key, value)

    def sign(self, method: str, endpoint: str | EndPoint, payload: dict[str, Any] | Sentinel = NotSet, *,
             timestamp: str = None, api_key: str = None) -> None:
        """Create HMAC signature for request

            Args
                method: HTTP method as a string
                endpoint: API endpoint path as a string or EndPoint enum instance
                payload: Dictionary containing API endpoint payload data.
                         This will be converted to a JSON string and hashed.
                         If no payload is provided, a default containing the currently set allyCode will be used.

            Keyword Args
                timestamp: Optional timestamp string to use instead of generating a new one. (primarily for testing)
                api_key: Optional API key to use instead of the one set in the container class. (primarily for testing)
        """
        # Removing 'api-key' header. This signals to the API that the request is HMAC signed
        if 'api-key' in self.headers:
            del self.headers['api-key']
            self.logger.debug("'api-key' header removed")

        # Get current unix timestamp in milliseconds
        if timestamp:
            req_time = timestamp
        else:
            req_time = str(int(time.time() * 1000))

        # Set the timestamp header for the API
        self.headers['x-timestamp'] = req_time
        self.logger.debug(f"'x-timestamp' header set to {self.headers['x-timestamp']}")

        # Create empty HMAC object using SHA256 algorithm
        if api_key:
            self.logger.debug(f"Using provided API key: [{self.get_api_key()}]")
            a_key = api_key.encode()
        else:
            self.logger.debug(f"Using API key from container class: [{self.get_api_key()}]")
            a_key = self.api_key.encode()
        hmac_obj = hmac.new(key=a_key, digestmod=hashlib.sha256)
        self.logger.debug(f"HMAC Hexdigest (base): {hmac_obj.hexdigest()}")

        # Add the timestamp to the HMAC object
        hmac_obj.update(req_time.encode())
        self.logger.debug(f"HMAC Hexdigest (timestamp): {hmac_obj.hexdigest()}")

        # Add the HTTP request method in all caps to the HMAC object
        hmac_obj.update(method.upper().encode())
        self.logger.debug(f"HMAC Hexdigest (HTTP method): {hmac_obj.hexdigest()}")

        # Add the API endpoint path to the HMAC object
        if isinstance(endpoint, EndPoint):
            endpoint = endpoint.value
        hmac_obj.update(endpoint.encode())
        self.logger.debug(f"HMAC Hexdigest (endpoint): {hmac_obj.hexdigest()}")

        # Create a serialized string of the payload object
        payload = self.payload if payload is NotSet else payload
        payload_str = dumps(payload, separators=(',', ':'))
        self.logger.debug(f"Payload string: {payload_str}")

        # Generate MD5 hash of the payload string
        payload_hash_digest = hashlib.md5(payload_str.encode()).hexdigest()
        self.logger.debug(f"Payload hash digest: {payload_hash_digest}")

        # Add the payload MD5 hash to the HMAC object
        hmac_obj.update(payload_hash_digest.encode())
        self.logger.debug(f"HMAC Hexdigest (payload): {hmac_obj.hexdigest()}")

        # Add the HMAC signature to the HTTP headers
        self.headers['Authorization'] = hmac_obj.hexdigest()
        self.client.headers = self.headers
        self.aclient.headers = self.headers
        self.logger.debug(f"HTTP client headers updated with HMAC signature: {self.client.headers}")



"""
Attribute definitions
"""

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any

__all__ = ["APIKey", "AllyCode", "Debug", "HMAC", "Headers", "Payload", "EndPoint"]


class ManagedAttribute(ABC):

    def __set_name__(self, owner, name):
        self.private_name = '_' + name

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.private_name)

    def __set__(self, obj, value):
        self.validate(value)
        setattr(obj, self.private_name, value)

    @abstractmethod
    def validate(self, value):
        pass


class APIKey(ManagedAttribute):

    def __init__(self, api_key=None):
        self.api_key = api_key

    def validate(self, value):
        if not isinstance(value, str):
            raise AttributeError(f"{value} must be a string, not type:{type(value)}")


class AllyCode(ManagedAttribute):

    def __init__(self, allycode=None):
        self.allycode = allycode

    def validate(self, value):
        if not isinstance(value, str):
            raise AttributeError(f"{value} must be a string, not type:{type(value)}")
        value = value.replace('-', '') if '-' in value else value


class DiscordId(ManagedAttribute):

    def __init__(self, allycode=None):
        self.allycode = allycode

    def validate(self, value):
        if not isinstance(value, str):
            raise AttributeError(f"{value} must be a string, not type:{type(value)}")
        value = value.replace('-', '') if '-' in value else value


class Debug(ManagedAttribute):

    def __init__(self, debug: bool = False):
        self.debug = debug

    def validate(self, value):
        if not isinstance(value, bool):
            raise AttributeError(f"{value} must be a boolean, not type:{type(value)}")


class HMAC(ManagedAttribute):

    def __init__(self, hmac: bool = True):
        self.hmac = hmac

    def validate(self, value):
        if not isinstance(value, bool):
            raise AttributeError(f"{value} must be a boolean, not type:{type(value)}")


class Headers(dict):

    def __getitem__(self, key):
        return super().get(key, None)

    def add_header(self, key: str, value: Any):
        """Add a header and append value if already exists"""
        if key in self:
            self[key] += f",{value}"
        else:
            self[key] = value

    push = add_header

    def delete_header(self, key: str):
        """Delete header if exists"""
        if key in self:
            del self[key]

    pop = delete_header


class Payload(dict):

    def __getitem__(self, key):
        return super().get(key, None)


class EndPoint(Enum):
    """Enum class for MBot API endpoints"""
    # API endpoints
    TW = 'tw'
    RAID = 'activeraid'
    TWLOGS = 'twlogs'
    TWLEADERBOARD = 'twleaderboard'
    GAC = "gac"
    INVENTORY = "inventory"
    TB = "tb"
    TBLOGS = "tblogs"
    EVENTS = "events"

    # Player Registry endpoints
    FETCH = "database"
    REGISTER = "comlink"
    VERIFY = "comlink"

    @staticmethod
    def get_endpoints():
        """Return a list of all endpoint names"""
        return [name for name, member in EndPoint.__members__.items()]
